/*  
    Conditional Temporal Median Filter

    Version 0.90  June 27, 2003

    Copyright (C) 2003 Kevin Atkinson (kevin.med@atkinson.dhs.org) under 
    the GNU GPL version 2.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "med.hpp"

#include <new>
#include <cstring>
#include <cstdlib>
#include <cassert>

#include <iostream>
#include <sstream>
#include <string>

#include <yuv4mpeg.h>

using std::cerr;
using std::cout;
using std::istringstream;
using std::memset;
using std::memcpy;
using std::string;

y4m_stream_info_t streaminfo;

size_t width, c_width, height, c_height;
size_t y_size, c_size, f_size;

struct Grid {
  byte * d;
  size_t w;
  byte & operator[](int i) {return d[i];}
  byte & operator()(size_t x, size_t y) {return d[w*y + x];}
  operator byte * () {return d;}
};

struct Border;

struct Borders {
  Border * data;
  Border * end;
  void setup(size_t);
  void clear() {end = data;}
};

// Frame data is stored in YUV 4:2:0 format where is component is stored
// in a seperate area of memory.
//   yuv[0] points to the y data
//   yuv[1] points to the u data which is 1/4 the size
//   yuv[2] points to the v data which is 1/4 the size
//
// This filter currently only modifies the y color component
//
// The 'y' grid points to the same data as yuv[0]
//
// The noise and motion grid are the same size as the y data

struct Frame {
  byte * yuv[3];
  Grid y;
  Grid noise;
  Grid motion;
  Borders borders;
  y4m_frame_info_t inf;
  byte * init(byte * buf);
};

void find_sizes(Borders & bds, Grid noise);
void mark_noise(Borders & bds, byte * noise);

typedef void Exec(byte * p, 
                  byte * c, byte * c_noise, byte * c_motion,
                  byte * n, byte * n_motion,
                  byte * o, size_t size);

extern "C" {
  void memzero(void *, size_t s);

  void find_motion(byte * p, byte * c, byte * motion, size_t);
  Exec cond_median;
  
  void motion_denoise(byte * moving, int width, int height);

  void find_outliers(byte * p, byte * c, byte * n, byte * o, size_t size);

  void mark_motion(byte * p, byte * p_noise,
                   byte * c, byte * c_noise, byte * c_motion,
                   size_t size);

  Exec remove_outliers;
  Exec mark_outliers;
  Exec map_outliers;
}

Exec * exec[2][3] = { {remove_outliers, mark_outliers, map_outliers},
                      {cond_median, map_outliers, map_outliers} };

//
//
//

byte P1 = 24;
byte P2 = 12;
//byte P1 = 32;
//byte P2 = 16;

short PWIDTH  = 16;
short PHEIGHT = 5;

int Y_NEXT = 2; // 1 if it is not interlaces

byte MTHRES = 16;
//byte MTHRES = 24;

int MWIDTH  = 7;
int MHEIGHT = 5;
byte MP = 11;

int MODE = M_SEARCH;
int SHOW = S_DENOISE;

byte MARK_V = 255;

//
//
//

void error(const char * err, const string & p = "")
{
  cerr << "Error: " << err;
  if (!p.empty())
    cerr << p;
  cerr << "\n";
  exit(1);
}

template <typename T>
void set_value(const string & p, T & var, const string & v)
{
  int val;
  istringstream s(v);
  s >> val;
  if (!s || val < 0) error("Invalid value for paramater ", p);
  var = val;
}

void help() {
#include "usage.cpp"
  exit(0);
}

int main(int argc, const char *argv[]) {

  for (int i = 1; i < argc; i += 2)
  {
    const char * p0 = argv[i];
    while (*p0 == '-') ++p0;
    string p(p0);
    if (p == "help")
      help();
    if (i + 1 >= argc)
      error("Value required for paramater ", p);
    string v(argv[i+1]);
    if (p == "p1")
      set_value(p, P1, v);
    else if (p == "p2")
      set_value(p, P2, v);
    else if (p == "pwidth")
      set_value(p, PWIDTH, v);
    else if (p == "pheight")
      set_value(p, PHEIGHT, v);
    else if (p == "interlaced") {
      if      (v == "false") Y_NEXT = 1;
      else if (v == "true")  Y_NEXT = 2;
      else error("Value for interlaced must be either 'true' or 'false'");} 
    else if (p == "mwidth")
      set_value(p, MWIDTH, v);
    else if (p == "mheight")
      set_value(p, MP, v);
    else if (p  == "mode") {
      if      (v == "search") MODE = M_SEARCH;
      else if (v == "median") MODE = M_MEDIAN;
      else error("Value for mode must be either 'search' or 'median'");}
    else if (p == "show") {
      if      (v == "denoise") SHOW = S_DENOISE;
      else if (v == "mark")    SHOW = S_MARK;
      else if (v == "map")     SHOW = S_MAP;
      else error("Value for show must be one of: denoise, mark, map");}
    else if (p == "mark_v")
      set_value(p, MARK_V, v);
    else 
      error("Invalid parameter: ", p);
  }

  y4m_init_stream_info(&streaminfo);
  if (y4m_read_stream_header (0, &streaminfo) != 0)
    error("Unable to read stream header from stdin");
  width = y4m_si_get_width (&streaminfo);
  height = y4m_si_get_height (&streaminfo);

  y4m_write_stream_header(1, &streaminfo);

  y_size = width*height;
  c_size = width*height/4;
  f_size = y_size * 3 + c_size * 2;

  c_height = height/2;
  c_width = width/2;

  byte buf[f_size * 5];
  memset(buf, 0, sizeof(buf));

  Frame frame[4];
  Frame fout;
  Frame * fp[4];

  byte * pos = buf;

  for (int i = 0; i != 4; ++i)
  {
    pos = frame[i].init(pos);
    fp[i] = frame + i;
  }
  pos = fout.init(pos);

  y4m_read_frame(0, &streaminfo, &fp[1]->inf, fp[1]->yuv);
  y4m_write_frame(1, &streaminfo, &fp[1]->inf, fp[1]->yuv);

  y4m_read_frame(0, &streaminfo, &fp[2]->inf, fp[2]->yuv);
  y4m_read_frame(0, &streaminfo, &fp[3]->inf, fp[3]->yuv);

  memset(fout.yuv[1], 0x80, c_size);
  memset(fout.yuv[2], 0x80, c_size);

  // prep
  if (MODE == M_SEARCH) {
    find_outliers(fp[1]->y, fp[2]->y, fp[3]->y, fp[2]->noise, y_size);
    find_sizes(fp[2]->borders, fp[2]->noise);
    mark_noise(fp[2]->borders, fp[2]->noise);
    mark_motion(fp[1]->y, fp[1]->noise, 
                fp[2]->y, fp[2]->noise, fp[2]->motion, y_size);
    motion_denoise(fp[2]->motion, width, height);
  } else {
    find_motion(fp[1]->y, fp[2]->y, fp[2]->motion, y_size);
    motion_denoise(fp[2]->motion, width, height);
  }

  byte tbuf1[y_size];
  byte tbuf2[y_size];

  int res = -1;
  do {
    //cerr << fp[0] << ' ' << fp[1] << ' ' << fp[2] << ' ' << fp[3] << '\n';
    // rotate
    Frame * ft = fp[0];
    for (int i = 0; i != 3; ++i) fp[i] = fp[i+1];
    fp[3] = ft;

    res = y4m_read_frame(0, &streaminfo, &fp[3]->inf, fp[3]->yuv);

    // prep
    if (MODE == M_SEARCH) {
      fp[2]->borders.clear();
      if (res == 0) {
        find_outliers(fp[1]->y, fp[2]->y, fp[3]->y, fp[2]->noise, y_size);
        find_sizes(fp[2]->borders, fp[2]->noise);
      }
      mark_noise(fp[2]->borders, fp[2]->noise);
      mark_motion(fp[1]->y, fp[1]->noise, 
                  fp[2]->y, fp[2]->noise, fp[2]->motion, y_size);
      motion_denoise(fp[2]->motion, width, height);
    } else {
      find_motion(fp[1]->y, fp[2]->y, fp[2]->motion, y_size);
      motion_denoise(fp[2]->motion, width, height);
    }
    
    (*exec[MODE][SHOW])(fp[0]->y,
                        fp[1]->y, fp[1]->noise, fp[1]->motion,
                        fp[2]->y, fp[2]->motion,
                        fout.y, y_size);

    if (SHOW != S_MAP) {
      fout.yuv[1] = fp[1]->yuv[1];
      fout.yuv[2] = fp[1]->yuv[2];
    }
    
    y4m_write_frame(1, &streaminfo, &fout.inf, fout.yuv);
    
  } while (res == 0);

  y4m_write_frame(1, &streaminfo, &fp[2]->inf, fp[2]->yuv);
}

//
// 
//

byte * Frame::init(byte * pos)
{
  y4m_init_frame_info(&inf);
  borders.setup(y_size);
  yuv[0] = pos;
  y.d     = pos;  y.w      = width; pos += y_size;
  noise.d = pos;  noise.w  = width; pos += y_size;
  motion.d = pos; motion.w = width; pos += y_size;
  yuv[1] = pos; pos += c_size;
  yuv[2] = pos; pos += c_size;
  return pos;
}

//
//
//

static const char B_NOTHING = 0;
static const char B_BORDER = 1;
static const char B_LINK = 2;
static const char B_SIZE = 3;
struct Border {
  char p1; // boolean
  char what;
  short ly;
  short lx1;
  short lx2;
  union Data {
    Data(short x = 0, short y = 0)
      {b.x1 = x; b.x2 = x; b.y1 = y; b.y2 = y;}
    struct {short x1; short x2; short y1; short y2;} b;
    struct {short width; short height; bool is_noise;} s;
    Border * nowis;
  } data;
  Border(short ly0, short lx10) : p1(false), what(B_BORDER),
                                  ly(ly0), lx1(lx10), lx2(lx10),
                                  data(lx10, ly0) {}
};

void Borders::setup(size_t s)
{
  data = (Border *) malloc (sizeof(Border) * (s/2 + 1));
  end = data;
}

//
//
//

void combine_xs(Borders & bds, Grid noise);
void combine_ys(Borders & bds);
void to_size(Borders & bds);

void find_sizes(Borders & bds, Grid noise)
{
  combine_xs(bds,noise);
  combine_ys(bds);
  to_size(bds);
}

void combine_xs(Borders & bds, Grid noise)
{
  byte * l = noise;
  Border * c = 0;
  for (size_t y = 0; y != height; ++y) 
  {
    for (size_t x = 0; x != width; ++x)
    {
      if (l[x] > P2) {

        if (c == 0) {c = bds.end++; new (c)Border(y,x);}
        if (l[x] > P1) c->p1 = true;

      } else if (c != 0) {

        c->data.b.x2 = x - 1;
        c->lx2 = x - 1; c = 0;

      }
    }
    if (c != 0) {c->lx2 = width - 1; c = 0;}
    l += width;
  }
}

void combine_ys(Borders & bds)
{
  bds.end->ly = height; // to avoid some special cases
  Border * c  = bds.data;
  Border * n = c + 1;
  while (n->ly == (n-1)->ly) ++n;
  Border * n0;

  bool c_has_next, n_has_next;
  
  while (n < bds.end) 
  {
    n0 = n;
    if (n->ly < c->ly + Y_NEXT) {++n; while (n->ly == (n-1)->ly) ++n;}

    if (c->ly + Y_NEXT == n->ly) {
      do {
        if ( (c->lx1 <= n->lx1 && c->lx2 >= n->lx1 - 1) ||
             (c->lx1 >  n->lx1 && n->lx2 >= c->lx1 - 1)   )
        {
          Border * a = c; while (a->what == B_LINK) a = a->data.nowis;
          Border * b = n; while (b->what == B_LINK) b = b->data.nowis;
          if (a > b) {Border * t = a; a = b; b = t;} // this is important
          a->p1 = a->p1 || b->p1;
          a->data.b.x1 = min(a->data.b.x1, b->data.b.x1);
          a->data.b.x2 = max(a->data.b.x2, b->data.b.x2);
          a->data.b.y2 = n->ly;
          if (a != b) {
            b->what = B_LINK;
            b->data.nowis = a;
          }
        }

        c_has_next = c->ly == (c+1)->ly;
        n_has_next = n->ly == (n+1)->ly;
        if      (c->lx2 <= n->lx2 && c_has_next) ++c;
        else if (n_has_next)                     ++n;
        else if (c_has_next)                     ++c;
        else break;

      } while (true);

      ++c;
      ++n;

    } else {
      if (c->ly < n0->ly && n0->ly < n->ly) {
        c = n0;
      } else {
        c = n;
        ++n;
        while (n->ly == (n-1)->ly) ++n;
      }
    }
  }
}

void to_size(Borders & bds)
{
  const short PH = PHEIGHT*Y_NEXT;
  for (Border * cur = bds.data; cur != bds.end; ++cur)
  {
    if (cur->what == B_BORDER) {
      Border::Data & b = cur->data;
      short w = b.b.x2 - b.b.x1 + 1;
      short h = b.b.y2 - b.b.y1 + 1;
      cur->what = B_SIZE; 
      b.s.is_noise = cur->p1 && w <= PWIDTH && h <= PH;
      b.s.width = w; 
      b.s.height = h;
    } else {
      //assert(cur->what == B_LINK);
      //assert(cur->data.nowis->what == B_SIZE);
      cur->what = B_SIZE;
      cur->data = cur->data.nowis->data;
    }
  }
}

void mark_noise(Borders & bds, byte * l)
{
  memzero(l, y_size);
  Border * c = bds.data;
  for (size_t y = 0; y != height; ++y)
  {
    for (; c->ly == y; ++c) {
      if (c->data.s.is_noise) {
        byte * b = l + c->lx1;
        byte * end = l + c->lx2 + 1;
        for (; b != end; ++b) *b = 0xFF;
      }
    }
    l += width;
  }
}

//
//
//
